package ee.ut.joshi.connectsix.ai.gametree;

import java.util.List;

import org.apache.log4j.Logger;

import ee.ut.joshi.connectsix.logic.factory.BoardCloneFactory;
import ee.ut.joshi.connectsix.main.ConnectObotProps;
import ee.ut.joshi.connectsix.model.Board;
import ee.ut.joshi.connectsix.model.Move;
import ee.ut.joshi.connectsix.model.Player;
import ee.ut.joshi.connectsix.model.player.MiniMaxPlayer;

/**
 * Standard MiniMax algorithm implementation. Full game tree is searched where
 * at each depth player tries to maximize the move value and opponent tries to
 * minimize the move value.
 * 
 * Board values are evaluated based on heuristics implementation.
 * 
 * Deprecated, use Negamax instead.
 */
@Deprecated
public class MiniMax {

	private static final Logger log = Logger.getLogger( MiniMax.class );

	/** Used to determine maximum search depth level */
	public static final int MAXDEPTH = ConnectObotProps.minimaxDepth();

	/** Singleton pattern */
	private static MiniMax instance = new MiniMax();

	private MiniMax() {
	}

	/**
	 * Singleton pattern - instantiation is restricted and there is only one
	 * instance of the MiniMax class.
	 * 
	 * @return Instance of MiniMax
	 */
	public static MiniMax getInstance() {

		return instance;
	}

	/**
	 * Starts recursive search in the game tree, looking for best moves for
	 * player and worst moves for player by opponent.
	 * 
	 * @param board Board state from where to start searching
	 * @param player Player who tries to maximize
	 * @param opponent Player who tries to minimize
	 * @return The best next move for player
	 */
	public Move minimax( Board<? extends Object> board, Player player, Player opponent ) {

		// Method follows same basics as minimaxMAX method, but
		// returns only Move instead of MoveScore Pair
		
		log.trace( "Entering minimax" );
		log.trace( "Players player=" + player + ";value=" + player.getValue() );
		log.trace( "Opponent=" + opponent + ";value=" + opponent.getValue() );
		log.trace( "Input board is " + board );

		List<Move> nextMoves = board.getNextMoves();
		
		// TODO: Make sure there is always at least one move here.
		
		log.trace( "List of next moves: " + nextMoves );

		MoveScorePair best = null;

		for( Move newmove : nextMoves ) {
			
			log.trace( "Looking for move: " + newmove );

			Board<? extends Object> cloneBoard = BoardCloneFactory.cloneBoard( board );

			cloneBoard.makeMove( player, newmove );
			
			log.trace( "Recursing further" );

			// Current method is for maximizing result, so next recursive call
			// would be for minimizing the score
			MoveScorePair msPair = minimaxMIN( cloneBoard, newmove, player, opponent, 0 );
			
			log.trace( "So far best move=" + best );
			log.trace( "Result from recursive call, msPair=" + msPair );

			if( best == null || best.score < msPair.score )
				best = msPair;
		}
		
		log.trace( "Found best move from recursive calls, best=" + best );
		
		return best.move;
	}

	private MoveScorePair minimaxMAX( Board<? extends Object> board, Move move, Player player, Player opponent,
			int depth ) {
		
		String dpth = "minimaxMAX-";
		
		for( int i = 0; i < depth; i++ ) {
			dpth += "-";
		}
		
		log.trace( dpth + "Entering minimaxMAX for move " + move );
		log.trace( dpth + "Input board is " + board );

		// Check if we reached max depth or game has ended already
		if( depth == MAXDEPTH || board.isGameEnd() ) {
			
			// Evaluate the board score
			Integer score = evaluateBoard( board, player );
			
			return new MoveScorePair( move, score );
		}

		// Find next moves
		List<Move> nextMoves = board.getNextMoves();
		
		log.trace( dpth + "Next possible moves=" + nextMoves );

		MoveScorePair best = null;

		// For each possible move we go recursively deeper down the game tree
		for( Move newmove : nextMoves ) {

			Board<? extends Object> cloneBoard = BoardCloneFactory.cloneBoard( board );

			cloneBoard.makeMove( player, newmove );

			// Recursive call to Minimize next move
			MoveScorePair msPair = minimaxMIN( cloneBoard, newmove, player, opponent, depth + 1 );

			log.trace( dpth + "So far best move=" + best );
			log.trace( dpth + "Result from recursive call, msPair=" + msPair );
			
			// If we find better move, store it
			if( best == null || best.score < msPair.score )
				best = msPair;
		}
		
		log.trace( dpth + "Found best move from recursive calls, best=" + best );

		// Return best found move
		return best;
	}

	private MoveScorePair minimaxMIN( Board<? extends Object> board, Move move, Player player, Player opponent,
			int depth ) {
		
		String dpth = "MINminimax-";
		
		for( int i = 0; i < depth; i++ ) {
			dpth += "-";
		}
		
		log.trace( dpth + "Entering minimaxMIN for move " + move );
		log.trace( dpth + "Input board is " + board );

		// Check if we reached max depth or game has ended already
		if( depth == MAXDEPTH || board.isGameEnd() ) {
			
			// Evaluate the board score
			Integer score = evaluateBoard( board, player );
			
			return new MoveScorePair( move, score );
		}

		// Find next moves
		List<Move> nextMoves = board.getNextMoves();
		
		log.trace( dpth + "Next possible moves=" + nextMoves );

		MoveScorePair worst = null;

		// For each possible move we go recursively deeper down the game tree
		for( Move newmove : nextMoves ) {

			Board<? extends Object> cloneBoard = BoardCloneFactory.cloneBoard( board );

			cloneBoard.makeMove( opponent, newmove );
			
			log.trace( dpth + "Recursing further" );

			// Recursive call to Minimize next move
			MoveScorePair msPair = minimaxMAX( cloneBoard, newmove, player, opponent, depth + 1 );

			log.trace( dpth + "So far worst move=" + worst );
			log.trace( dpth + "Result from recursive call, msPair=" + msPair );
			
			// If we find better move, store it
			if( worst == null || worst.score > msPair.score )
				worst = msPair;
		}

		log.trace( dpth + "Found worst move from recursive calls, worst=" + worst );

		// Return best found move
		return worst;
	}

	private Integer evaluateBoard( Board<? extends Object> board, Player player ) {

		MiniMaxPlayer mmPlr = (MiniMaxPlayer) player;

		return mmPlr.evaluateBoard( board );
	}

	/**
	 * Helper class to store moves and their score value within game tree search
	 * 
	 */
	private class MoveScorePair implements Comparable<MoveScorePair> {

		private Move move;
		private Integer score;

		private MoveScorePair( Move move, Integer score ) {

			this.move = move;
			this.score = score;
		}

		@Override
		public int compareTo( MoveScorePair msPair ) {

			return score.compareTo( msPair.score );
		}

		@Override
		public String toString() {
			return "[move=" + move + ", score=" + score + "]";
		}
	}
}
