package ee.ut.joshi.connectsix.ai.gametree;

import java.util.List;

import org.apache.log4j.Logger;

import ee.ut.joshi.connectsix.logic.factory.BoardCloneFactory;
import ee.ut.joshi.connectsix.main.ConnectObotProps;
import ee.ut.joshi.connectsix.model.Board;
import ee.ut.joshi.connectsix.model.Move;
import ee.ut.joshi.connectsix.model.Player;
import ee.ut.joshi.connectsix.model.player.NegaMaxPlayer;

//@formatter:off
//PSEUDOCODE:
//	function negamax(node, depth, a, b, color)
//    if node is a terminal node or depth = 0
//        return color * the heuristic value of node
//    else
//        foreach child of node
//            val := -negamax(child, depth-1, -b, -a, -color)
//            {the following if statement constitutes alpha-beta pruning}
//            if val=a
//                return val
//            if val=a
//                a:=val
//        return a
//@formatter:on
public class Negamax {

	private static final Logger log = Logger.getLogger( Negamax.class );

	/** Used to determine maximum search depth level */
	public static int MAXDEPTH = ConnectObotProps.minimaxDepth();

	protected boolean printMoveScores = false;

	// Integer.MAX_VALUE sux, big time.
	public static final int MAXVALUE = 1000000;
	public static final int MINVALUE = -1000000;

	private Player player;
	private Player opponent;

	public Negamax( Player plr, Player opp ) {

		player = plr;
		opponent = opp;
	}

	public Move negamax( Board<? extends Object> board ) {

		log.debug( "Entering negamax" );
		log.debug( "Players player=" + player + ";value=" + player.getValue() );
		log.debug( "Opponent=" + opponent + ";value=" + opponent.getValue() );
		log.debug( "Input board is:" );
		logBoard( board );

		// Find next moves
		List<Move> moves = board.getNextMoves();
		log.trace( "Next possible moves=" + moves );

		// TODO: Make sure there is always at least one move here.

		MoveScorePair best = null;

		// log.trace( "Taking maximum of these: " );

		for( Move newmove : moves ) {

			// log.trace( "Looking for move: " + newmove );

			Board<? extends Object> cloneBoard = BoardCloneFactory.cloneBoard( board );

			cloneBoard.makeMove( player, newmove );

			// Current method is for maximizing result, so next recursive call
			// would be for minimizing the score
			Integer value = -negamaxRec( cloneBoard, player, 0, MINVALUE, MAXVALUE, -1 );

			MoveScorePair msPair = new MoveScorePair( newmove, value );

			// log.trace( "So far best move=" + best );
			log.debug( "Result from recursive call, msPair=" + msPair );
			if( printMoveScores )
				Logger.getLogger( "playermoves" ).trace( "Result from recursive call, msPair=" + msPair );

			if( best == null || best.getScore() <= msPair.getScore() )
				best = msPair;
		}

		log.trace( "Found best move from recursive calls, best=" + best );

		return best.getMove();
	}

	protected Integer negamaxRec( Board<? extends Object> board, Player plr, int depth, int alpha, int beta, int color ) {

		String dpth = "negamax-" + plr + "-";

		for( int i = 0; i < depth; i++ ) {
			dpth += "-";
		}

		// log.trace( dpth + "Input board is " + board );

		// Check if game is over and evaluate with min/max values
		if( board.isGameEnd() ) {

			log.trace( dpth + "Detected game end plr=" + board.getWinningPlayer() + ";depth=" + depth + ";color=" + color );

			if( board.getWinningPlayer() == player.getValue() )
				return color * MAXVALUE;

			if( board.getWinningPlayer() == opponent.getValue() )
				return color * MINVALUE;

			if( board.getWinningPlayer() == Board.EMPTY )
				return color * evaluateBoard( board, plr );
		}

		// Check if we reached max depth so we can evaluate the board
		if( depth == MAXDEPTH ) {

			// Evaluate the board score
			Integer score = evaluateBoard( board, plr );

			return color * score;
		}

		// Find next moves
		List<Move> moves = board.getNextMoves();
		log.trace( dpth + "Next possible moves=" + moves );

		// String minmax = color == 1 ? "maximum" : "minimum";
		// log.trace( dpth + "Taking " + minmax + " of these:" );

		// For each possible move we go recursively deeper down the game tree
		for( Move newmove : moves ) {

			Board<? extends Object> cloneBoard = BoardCloneFactory.cloneBoard( board );

			Player nextPlr = plr == player ? opponent : player;

			cloneBoard.makeMove( nextPlr, newmove );

			// log.trace( dpth + "Recursing further for move=" + newmove );

			// Recursive call to Minimize next move
			Integer value = -1 * negamaxRec( cloneBoard, nextPlr, depth + 1, -beta, -alpha, -color );

			// log.trace( dpth + "So far alpha=\t" + alpha + ";beta=\t" + beta );
			log.trace( dpth + "Result from recursive call, value=" + value + ";move=" + newmove );

			if( value >= beta ) {

				log.trace( dpth + "Performing beta cutoff, value=" + value + ";beta=" + beta );

				return value;
			}

			if( value >= alpha ) {

				log.trace( dpth + "Found better alpha, value=" + value + ";alpha=" + alpha );

				alpha = value;
			}
		}

		return alpha;
	}

	protected Integer evaluateBoard( Board<? extends Object> board, Player player ) {

		NegaMaxPlayer mmPlr = (NegaMaxPlayer) player;

		return mmPlr.evaluateBoard( board );
	}

	public void printMoveScores() {

		printMoveScores = true;
	}

	protected void logBoard( Board<? extends Object> board ) {

		String brdStr = board.toString();

		for( String row : brdStr.split( "\n" ) ) {

			log.debug( row );
		}
	}
}
