package ee.ut.joshi.connectsix.ai.gametree;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.apache.log4j.Logger;

import ee.ut.joshi.connectsix.logic.board.BoardComparator;
import ee.ut.joshi.connectsix.logic.factory.BoardCloneFactory;
import ee.ut.joshi.connectsix.model.Board;
import ee.ut.joshi.connectsix.model.Move;
import ee.ut.joshi.connectsix.model.Player;

//@formatter:off
//PSEUDOCODE:
//	function negamax(node, depth, a, b, color)
//  if node is a terminal node or depth = 0
//      return color * the heuristic value of node
//  else
//      foreach child of node
//          val := -negamax(child, depth-1, -b, -a, -color)
//          {the following if statement constitutes alpha-beta pruning}
//          if val=a
//              return val
//          if val=a
//              a:=val
//      return a
//@formatter:on
public class NegamaxV3 extends Negamax {

	private static final Logger log = Logger.getLogger( NegamaxV3.class );

	private Player player;
	private Player opponent;

	public NegamaxV3( Player plr, Player opp ) {

		super( plr, opp );

		player = plr;
		opponent = opp;
	}

	@Override
	public Move negamax( Board<? extends Object> board ) {
		// negamax( board, 0, alpha=MINVALUE, beta=MAXVALUE 1)

		if( log.isTraceEnabled() ) {

			log.trace( "Entering negamax" );
			log.trace( "Players player=" + player + ";value=" + player.getValue() );
			log.trace( "Opponent=" + opponent + ";value=" + opponent.getValue() );
			log.trace( "Input board is:" );
			logBoard( board );
		}

		int alpha = MINVALUE;
		int beta = MAXVALUE;

		List<Move> moves = board.getNextMoves();
		log.trace( "Next possible moves=" + moves );

		MoveScorePair best = new MoveScorePair( moves.get( 0 ), MINVALUE );

		List<Board<? extends Object>> cloneboards = new ArrayList<>();

		for( Move newmove : moves ) {

			Board<? extends Object> cloneBoard = BoardCloneFactory.cloneBoard( board );
			cloneBoard.makeMove( player, newmove );

			cloneboards.add( cloneBoard );
		}

		Collections.sort( cloneboards, Collections.reverseOrder( BoardComparator.instance() ) );

		for( Board<? extends Object> cloneboard : cloneboards ) {

			// Integer value = -negamaxRec( cloneBoard, player, 0, MINVALUE, MAXVALUE, -1 );
			Integer value = -negamaxRec( cloneboard, player, 0, -beta, -alpha, -1 );

			MoveScorePair msPair = new MoveScorePair( cloneboard.getLastMove(), value );

			if( printMoveScores )
				Logger.getLogger( "playermoves" ).trace( "Result from recursive call, msPair=" + msPair );

			if( value > best.getScore() )
				best = msPair;

			// Can only happen if we find a guaranteed win
			if( value >= beta ) {

				log.trace( "Performing beta cutoff, value=" + value + ";beta=" + beta );

				return best.getMove();
			}

			if( value > alpha ) {

				log.trace( "Found better alpha, value=" + value + ";alpha=" + alpha );

				alpha = value;
			}
		}

		return best.getMove();
	}

	@Override
	public Integer negamaxRec( Board<? extends Object> board, Player plr, int depth, int alpha, int beta, int color ) {

		if( board.isGameEnd() ) {

			if( board.getWinningPlayer() == player.getValue() )
				return color * MAXVALUE;

			if( board.getWinningPlayer() == opponent.getValue() )
				return color * MINVALUE;

			return color * evaluateBoard( board, plr );
		}

		if( depth == MAXDEPTH )
			return color * evaluateBoard( board, plr );

		Player nextPlr = plr == player ? opponent : player;

		String dpth = "negamax-" + nextPlr + "-";

		for( int i = 0; i < depth; i++ ) {
			dpth += "-";
		}

		List<Move> moves = board.getNextMoves();

		if( log.isTraceEnabled() )
			log.trace( dpth + "Next possible moves=" + moves );

		List<Board<? extends Object>> cloneboards = new ArrayList<>();

		for( Move newmove : moves ) {

			Board<? extends Object> cloneBoard = BoardCloneFactory.cloneBoard( board );
			cloneBoard.makeMove( nextPlr, newmove );

			cloneboards.add( cloneBoard );
		}

		if( color == 1 && depth != MAXDEPTH - 1 )
			Collections.sort( cloneboards, Collections.reverseOrder( BoardComparator.instance() ) );
		else if( color == -1 && depth != MAXDEPTH - 1 )
			Collections.sort( cloneboards, BoardComparator.instance() );

		for( Board<? extends Object> cloneBoard : cloneboards ) {

			Integer value = -1 * negamaxRec( cloneBoard, nextPlr, depth + 1, -beta, -alpha, -color );

			if( log.isTraceEnabled() )
				log.trace( dpth + "Result from recursive call, value=" + value + ";move=" + cloneBoard.getLastMove() );

			if( value >= beta ) {

				log.trace( dpth + "Performing beta cutoff, value=" + value + ";beta=" + beta );

				return value;
			}

			if( value >= alpha ) {

				log.trace( dpth + "Found better alpha, value=" + value + ";alpha=" + alpha );

				alpha = value;
			}
		}

		return alpha;
	}
}
