package ee.ut.joshi.connectsix.ai.gametree.threaded;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.RejectedExecutionException;

import org.apache.log4j.Logger;

import ee.ut.joshi.connectsix.ai.gametree.MoveScorePair;
import ee.ut.joshi.connectsix.ai.gametree.NegamaxV3;
import ee.ut.joshi.connectsix.logic.board.BoardComparator;
import ee.ut.joshi.connectsix.logic.factory.BoardCloneFactory;
import ee.ut.joshi.connectsix.main.ConnectObotProps;
import ee.ut.joshi.connectsix.model.Board;
import ee.ut.joshi.connectsix.model.Move;
import ee.ut.joshi.connectsix.model.Player;

public class ThreadedNegamax extends NegamaxV3 {

	private static final Logger log = Logger.getLogger( ThreadedNegamax.class );

	public static final int NROFTHREADS = ConnectObotProps.nrOfThreads();

	private final ExecutorService execSvc = Executors.newFixedThreadPool( NROFTHREADS );

	private List<NegamaxThread> allThreads = new ArrayList<>();
	private List<NegamaxThread> queueThreads = new ArrayList<>();
	private List<NegamaxThread> runningThreads = new ArrayList<>();

	private Integer bestAlpha = MINVALUE;
	private Integer bestBeta = MAXVALUE;

	private NegamaxManager manager;

	private Player player;
	private Player opponent;

	MoveScorePair bestMove;

	public ThreadedNegamax( NegamaxManager manager, Player plr, Player opp ) {

		super( plr, opp );

		this.manager = manager;

		player = plr;
		opponent = opp;
	}

	@Override
	public Move negamax( Board<? extends Object> board ) {

		if( log.isTraceEnabled() ) {

			log.trace( "Entering threaded negamax" );
			log.trace( "Players player=" + player + ";value=" + player.getValue() );
			log.trace( "Opponent=" + opponent + ";value=" + opponent.getValue() );
			log.trace( "Input board is:" );
			logBoard( board );
		}

		List<Move> moves = board.getNextMoves();
		log.trace( "Next possible moves=" + moves );

		bestMove = new MoveScorePair( moves.get( 0 ), MINVALUE );

		List<Board<? extends Object>> cloneboards = new ArrayList<>();

		for( Move newmove : moves ) {

			Board<? extends Object> cloneBoard = BoardCloneFactory.cloneBoard( board );
			cloneBoard.makeMove( player, newmove );

			cloneboards.add( cloneBoard );
		}

		Collections.sort( cloneboards, Collections.reverseOrder( BoardComparator.instance() ) );

		for( Board<? extends Object> cloneboard : cloneboards ) {

			// Integer value = -negamaxRec( cloneboard, player, 0, -beta, -alpha, -1 );

			NegamaxThread thread = new NegamaxThread( this, player, opponent );

			thread.setAlpha( -MAXVALUE ); // Same as -beta
			// Beta should be configured in start of thread process with current best alpha value
			// thread.setBeta( -getBestAlpha() ); // Same as -alpha
			thread.setBoard( cloneboard );
			thread.setColor( -1 );
			thread.setDepth( 0 );
			thread.setPlr( player );

			allThreads.add( thread );
			queueThreads.add( thread );
		}

		startNegamax();

		// Notify NegamaxManager that all threads are finished are ready to be used again
		manager.setThreadedDeactivated();

		log.debug( "Returning bestMove=" + bestMove );

		return bestMove.getMove();
	}

	private void startNegamax() {

		// log.debug( "Starting theaded negamax, queuedThreads=" + queueThreads.size() );

		boolean allComplete = false;

		while( !allComplete && !execSvc.isTerminated() ) {

			// Assume we are done until proven otherwise
			allComplete = true;

			activateNewThreads();

			for( NegamaxThread thread : allThreads ) {

				// If we find a non completed thread, we set the completion status to false;
				allComplete = !thread.isCompleted() ? false : allComplete;
			}

			try {
				Thread.sleep( 100 );
			} catch( InterruptedException e ) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

		execSvc.shutdownNow();

		// log.debug( "Threaded negamax completed." );
	}

	private synchronized void activateNewThreads() {

		Integer maxThreads = manager.maxUsableThreads();

		while( runningThreads.size() < maxThreads && !execSvc.isShutdown() ) {

			if( queueThreads.size() == 0 )
				return; // No more threads to run

			// log.debug( "Starting new NegamaxThread, current queue size=" + queueThreads.size() );

			NegamaxThread thread = queueThreads.get( 0 );

			thread.setBeta( -getBestAlpha() ); // Same as -alpha

			try {

				execSvc.execute( thread );

			} catch( RejectedExecutionException e ) {

				log.trace( "Execution of minimax thread rejected, message=" + e.getMessage() );

				return;
			}

			runningThreads.add( thread );
			queueThreads.remove( thread );
		}
	}

	public synchronized void reportResult( NegamaxThread thread ) {

		// If we found beta cutoff, we can't accept new results.
		if( execSvc.isShutdown() )
			return;

		// log.debug( "A thread is reporting with new result" );

		Integer value = thread.getResultValue();
		Move move = thread.getBoard().getLastMove();

		// log.debug( "Removing thread from runningThreads=" + runningThreads.size() );
		runningThreads.remove( thread );
		// log.debug( "Thread removed, new queue size=" + runningThreads.size() );

		if( printMoveScores ) {
			Logger.getLogger( "playermoves" ).trace( "Result from recursive call, msPair=" + new MoveScorePair( move, value ) );

			// TODO: Add 'progress bar'
			// System.out.print( "." );
			// System.out.flush();
		}

		// Can only happen if we find a guaranteed win
		if( value >= bestBeta ) {

			// log.debug( "Performing beta cutoff, beta=" + bestBeta + ";value=" + value );

			bestMove = new MoveScorePair( move, value );

			execSvc.shutdownNow();

			return;
		}

		if( value > getBestAlpha() ) {

			// log.debug( "Found better alpha, value=" + value + ";alpha=" + getBestAlpha() );

			bestMove = new MoveScorePair( move, value );

			this.bestAlpha = value;
		}

		activateNewThreads();
	}

	public synchronized Integer getBestAlpha() {
		return bestAlpha;
	}

	@Override
	public Integer evaluateBoard( Board<? extends Object> board, Player player ) {
		// Have to make public so thread can access here
		return super.evaluateBoard( board, player );
	}
}
