package ee.ut.joshi.connectsix.ai.genetic;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import ee.ut.joshi.connectsix.ai.randomforest.train.TrainForest;
import ee.ut.joshi.connectsix.main.ConnectObotProps;

/**
 * Genetic evolution algorithm process manager - controls the process of
 * evolution in each iteration with help of Knowledge
 */
public class Brain {

	private static final Logger log = Logger.getLogger( Brain.class );

	/** Initial and iteration end population size */
	public static final int POPULATION = ConnectObotProps.geneticPopulation();

	/** Maximum population size after mating process */
	public static final int MAXPOP = POPULATION * 2;

	/** Mutation ratio after each new individual creation */
	public static final double MUTATION = ConnectObotProps.geneticMutation();

	/** Used to train new Random Forests */
	private TrainForest trainForest;

	/** Genetic algorithm helper class, e.g. for mating process @HasGetter @HasSetter */
	private Knowledge knowledge;

	/** All individuals including child individuals after mating process */
	private List<Individual> population = new ArrayList<>( MAXPOP );

	/**
	 * Initializes brain for genetic algorithm.
	 * 
	 * @param trainForest Used to train new forests and trees
	 */
	public Brain( TrainForest trainForest ) {

		this.trainForest = trainForest;
		knowledge = new Knowledge( trainForest );
	}

	/**
	 * Initializes starting population.
	 */
	public void initPopulation() {

		log.debug( "Initializing population." );

		InitPopulation initpop = new InitPopulation( trainForest );
		initpop.initPopulation( population );
	}

	/**
	 * Iteration step of genetic algorithm:<br>
	 * - individuals are sorted by their value<br>
	 * - best individuals are kept for next population seed<br>
	 * - mates best individuals to produce new child individuals<br>
	 * - mutates individuals based on mutation ratio<br>
	 * <br>
	 * Population size at start and end of the iteration is always MAXPOP<br>
	 */
	public void iteration() {

		ArrayList<Individual> newPop = new ArrayList<>( MAXPOP );

		Individual child;

		// Find best individuals
		population = knowledge.naturalSelection( population );

		// Retrain population on new data but on same patterns
		knowledge.retrainPopulation( population );

		// Mate best individuals

		// for( Individual father : population ) {
		// for( Individual mother : population ) {
		while( newPop.size() < MAXPOP - POPULATION ) {

			Individual father = knowledge.chooseMate( population );
			Individual mother = knowledge.chooseMate( population );

			if( father == mother )
				continue;

			child = knowledge.mate( father, mother );

			// Mutate individual
			if( Math.random() < MUTATION )
				knowledge.mutate( child );

			newPop.add( child );
		}

		population.addAll( newPop );
	}

	public Knowledge getKnowledge() {
		return knowledge;
	}

	public void setKnowledge( Knowledge knowledge ) {
		this.knowledge = knowledge;
	}

	public List<Individual> getPopulation() {
		return population;
	}

	public void setPopulation( ArrayList<Individual> population ) {
		this.population = population;
	}
}
