package ee.ut.joshi.connectsix.ai.genetic;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Random;

import org.apache.log4j.Logger;

import ee.ut.joshi.connectsix.ai.genetic.genepool.GeneTrainer;
import ee.ut.joshi.connectsix.ai.genetic.genepool.IndividualRetrainer;
import ee.ut.joshi.connectsix.ai.randomforest.RandomForest;
import ee.ut.joshi.connectsix.ai.randomforest.RandomWoods;
import ee.ut.joshi.connectsix.ai.randomforest.train.TrainForest;

/**
 * Genetic algorithm process center<br>
 * - mates two individuals by combining their genetic information<br>
 * - makes natural selection by discarding bad individuals and keeping best ones<br>
 * - mutates single individual<br>
 */
public class Knowledge {

	private static final Logger log = Logger.getLogger( Knowledge.class );

	/** Used to generate random numbers */
	private final Random random = new Random();

	/** Used to train new trees in mutation process */
	private final GeneTrainer geneTrainer;

	/** Used to train new Random Forests and Woods */
	private TrainForest trainForest;

	/**
	 * Initializes knowledge to help Brain in genetic algorithm.
	 * 
	 * @param trainForest Used to train new trees in mutation process
	 */
	protected Knowledge( TrainForest trainForest ) {

		geneTrainer = new GeneTrainer( trainForest );
		this.trainForest = trainForest;
	}

	/**
	 * Used to select a individual for mating process. For example if algorithm is implemented to use not all, but some portion of the
	 * population in mating process in order to keep the maximum population smaller.
	 * 
	 * @param population Population from where to choose the mate
	 * @return Returns a single individual for mating process
	 */
	protected Individual chooseMate( List<Individual> population ) {

		return population.get( random.nextInt( Brain.POPULATION ) );
	}

	/**
	 * Core of the genetic algorithm where parent genetic information is transferred to child individual.<br>
	 * <br>
	 * Note: It might happen that child individual will have duplicate genes.
	 * 
	 * @param father One of the individuals to mate
	 * @param mother Second of the individuals to mate
	 * @return New child individual combined from input individuals
	 */
	protected Individual mate( Individual father, Individual mother ) {

		RandomForest childForest = new RandomForest();

		// Resulting child individual
		Individual child = new Individual( childForest );

		// Parent individual genetic information
		List<RandomWoods> fatherWoods = father.getForest().getWoods();
		List<RandomWoods> motherWoods = mother.getForest().getWoods();

		List<RandomWoods> childWoods = new ArrayList<>( TrainForest.nrOfTrees );

		// Core of mating - either take a decision tree from father or mother
		// and transfer it to child individual
		for( int i = 0; i < TrainForest.nrOfTrees; i++ ) {

			if( random.nextDouble() < 0.5 ) {

				childWoods.add( fatherWoods.get( i ) );

			} else {

				childWoods.add( motherWoods.get( i ) );
			}
		}

		// TODO: Add to properties, this is for removing duplicate trees
		// for( int i = 0; i < TrainForest.nrOfTrees; i++ ) {
		// for( int j = 0; j < TrainForest.nrOfTrees; j++ ) {
		//
		// if( i == j )
		// continue;
		//
		// // Replace identical trees with new one
		// if( childWoods.get( i ) == childWoods.get( j ) )
		// childWoods.set( i, geneTrainer.getForestWoods() );
		// }
		// }

		// Attach genetic information data to child individual
		childForest.setWoods( childWoods );

		return child;
	}

	/**
	 * Mutates a forest by replacing individual trees with completely new fully trained tree. Determined by MUTATION factor.
	 * 
	 * @param child Returns mutated individual
	 */
	protected void mutate( Individual child ) {

		List<RandomWoods> woods = child.getForest().getWoods();
		List<RandomWoods> addList = new ArrayList<>();
		List<RandomWoods> removeList = new ArrayList<>();

		// Mutation core - replace some woods with new woods
		for( RandomWoods fWoods : woods ) {

			if( random.nextDouble() < Brain.MUTATION ) {

				// addList.add( trainForest.trainForestTree() );
				addList.add( geneTrainer.getForestWoods() );
				removeList.add( fWoods );
			}
		}

		int nrOfTrees = woods.size();

		// In case individual has duplicate genes, replace one of the genes
		// with a new trained decision tree
		for( int i = 0; i < nrOfTrees - 1; i++ ) {
			for( int j = i + 1; j < nrOfTrees; j++ ) {

				if( woods.get( i ) == woods.get( j ) )
					woods.set( i, geneTrainer.getForestWoods() );
			}
		}

		woods.addAll( addList );
		woods.removeAll( removeList );

		// In case for some reason we are still missing correct number of trees.
		while( woods.size() < nrOfTrees ) {

			// trees.add( trainForest.trainForestTree() );
			woods.add( geneTrainer.getForestWoods() );
		}
	}

	/**
	 * Retrains all individuals in a population with new acquired data from previous iteration, but with same pattern.
	 * 
	 * @param population Population to be retrained
	 */
	protected void retrainPopulation( List<Individual> population ) {

		IndividualRetrainer retrainer = new IndividualRetrainer( trainForest );

		for( Individual individual : population ) {

			retrainer.retrainIndividual( individual );
		}

		retrainer.startRetraining();
	}

	/**
	 * Sorts all individuals by their fitness value and returns the best ones for next generation training.
	 * 
	 * @param oldPop Full population of individuals
	 * @return Population of best individuals
	 */
	protected List<Individual> naturalSelection( List<Individual> oldPop ) {

		log.debug( "Starting natural selection." );

		String debug = "";
		if( log.isDebugEnabled() ) {
			for( Individual individual : oldPop ) {
				debug += individual.getValue() + ";";
			}
			log.debug( "Values of individuals(" + oldPop.size() + "): " + debug );
		}

		Collections.sort( oldPop, new IndividualComparator() );
		Collections.reverse( oldPop );

		// newPop = newPop.subList( 0, Brain.POPULATION );

		List<Individual> newPop = new ArrayList<Individual>();

		for( int i = 0; i < Brain.POPULATION; i++ ) {

			newPop.add( oldPop.get( i ) );
		}

		for( int i = 0; i < newPop.size(); i++ ) {

			PopulationStats.logForest( i, newPop.get( i ) );
		}

		PopulationStats.logStats( newPop );
		PopulationStats.logGenerations( newPop );
		PopulationStats.logGamestats( newPop );

		for( Individual individual : newPop ) {
			individual.setValue( 0 );
			individual.incGeneration();
			individual.setWinsOne( 0 );
			individual.setDrwsOne( 0 );
			individual.setLossOne( 0 );
			individual.setWinsTwo( 0 );
			individual.setDrwsTwo( 0 );
			individual.setLossTwo( 0 );
		}

		debug = "";
		if( log.isTraceEnabled() ) {
			for( Individual individual : newPop ) {
				debug += individual.getValue() + ";";
			}
			log.trace( "Reset values of individuals: " + debug );
		}

		return newPop;
	}
}
