package ee.ut.joshi.connectsix.ai.randomforest;

import java.io.BufferedWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

import ee.ut.joshi.connectsix.ai.randomforest.dataattr.DataAttribute;
import ee.ut.joshi.connectsix.ai.randomforest.dataattr.DataEntry;

/**
 * Random forest is an ensemble classifier that consists of many decision trees and outputs the class that is the mode of the classes output
 * by individual trees.
 */
public class RandomWoods {

	private static final Logger log = Logger.getLogger( RandomWoods.class );

	/** Stores the base attributes that the woods was trained on */
	private List<Integer> baseAttrs = new ArrayList<>();

	/** Individual decision trees */
	private List<ForestTree> trees = new ArrayList<>();

	/**
	 * @param tree A decision tree to be added into forest
	 */
	public void addTree( ForestTree tree ) {

		trees.add( tree );
	}

	/**
	 * Feeds input to each tree in forest and collects the category classification from them. Result is most frequent category as decided by
	 * decision trees.
	 * 
	 * @param dataEntry Data entry on which to find a category
	 * @return Most frequent category decided by decision trees
	 */
	public DataAttribute findCategory( DataEntry dataEntry ) {

		log.trace( "Looking category for\n" + dataEntry );

		Set<Map.Entry<DataAttribute, Integer>> mapSet = findCategoryMap( dataEntry ).entrySet();

		Map.Entry<DataAttribute, Integer> better = mapSet.iterator().next();

		// Find most frequent attribute
		for( Map.Entry<DataAttribute, Integer> entry : mapSet ) {

			if( entry.getValue() > better.getValue() )
				better = entry;
		}

		log.trace( "Found category for\n" + dataEntry );
		log.trace( "Most frequent category=" + better.getKey() + ";count=" + better.getValue() );
		log.trace( "dataEntry category=" + dataEntry.getCategory() + "; found=" + better.getKey() );

		return better.getKey();
	}

	/**
	 * Creates a map where keys are category names and value is count of the category classifications by all trees.
	 * 
	 * @param dataEntry Data entry on which to find a category
	 * @return Map with category frequency count
	 */
	public Map<DataAttribute, Integer> findCategoryMap( DataEntry dataEntry ) {

		Map<DataAttribute, Integer> resultMap = new HashMap<DataAttribute, Integer>();

		for( ForestTree tree : trees ) {

			DataAttribute temp = tree.findCategory( dataEntry );

			if( resultMap.containsKey( temp ) ) {

				int count = resultMap.get( temp );

				resultMap.put( temp, count + 1 );

			} else {

				resultMap.put( temp, 1 );
			}
		}

		return resultMap;
	}

	public List<Integer> getBaseAttrs() {
		return baseAttrs;
	}

	public void setBaseAttrs( List<Integer> baseAttrs ) {
		this.baseAttrs = baseAttrs;
	}

	public List<ForestTree> getTrees() {
		return trees;
	}

	public void setTrees( List<ForestTree> trees ) {
		this.trees = trees;
	}

	@Override
	public String toString() {

		String result = "[RandomWoods]\n";

		for( ForestTree tree : trees ) {

			result += tree;
		}

		return result;
	}

	public String toStringXML() {

		String result = "\n\t<RandomWoods>";

		for( ForestTree tree : trees ) {

			result += tree.toStringXML();
		}

		result += "\n\t</RandomWoods>";

		return result;
	}

	public void toStringXML( BufferedWriter out ) throws IOException {

		out.write( "\n\t<RandomWoods>" );

		for( ForestTree tree : trees ) {

			tree.toStringXML( out );
		}

		out.write( "\n\t</RandomWoods>" );
	}
}
