package ee.ut.joshi.connectsix.ai.randomforest.train;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.apache.log4j.Logger;

import ee.ut.joshi.connectsix.ai.randomforest.dataattr.DataAttribute;
import ee.ut.joshi.connectsix.ai.randomforest.dataattr.DataEntry;
import ee.ut.joshi.connectsix.ai.randomforest.dataattr.DataSet;
import ee.ut.joshi.connectsix.ai.randomforest.db.DataSetFactory;

/**
 * Gain measures how well a given attribute separates training examples into
 * targeted classes. The one with the highest information (information being the
 * most useful for classification) is selected.
 */
public class TrainGain {

	private static final Logger log = Logger.getLogger( TrainGain.class );

	private TrainEntropy ent = new TrainEntropy();

	/**
	 * Calculates gain value of an attribute within given data set.
	 * 
	 * @param dataSet Data set to be analyzed.
	 * @param attrIndx Attribute index to be measured
	 * @return Gain value of attribute within data set
	 */
	public Double calculateGain( DataSet dataSet, Integer attrIndx ) {

		log.trace( "Calculating gain" );

		// Gain(S, A) = Entropy(S) - S ((|Sv| / |S|) * Entropy(Sv))
		// S is each value v of all possible values of attribute A
		// Sv = subset of S for which attribute A has value v
		// |Sv| = number of elements in Sv
		// |S| = number of elements in S

		Double result = 0.00;

		// Entropy(S)
		result = ent.calcEntropy( dataSet );

		HashMap<DataAttribute, Integer> map = new HashMap<>();

		// Find frequency for attribute values in the data set
		for( DataEntry entry : dataSet.getData() ) {

			DataAttribute attrVal = entry.getAttrValue( attrIndx );

			if( map.containsKey( attrVal ) ) {

				Integer count = map.get( attrVal );

				map.put( attrVal, count + 1 );

			} else {

				map.put( attrVal, 1 );
			}
		}

		// S ((|Sv| / |S|) * Entropy(Sv))
		for( DataAttribute key : map.keySet() ) {

			// (|Sv| / |S|) * Entropy(Sv)
			result = result - subsetEntropy( dataSet, attrIndx, key );
		}

		log.trace( "Gain is for attrIndx=" + attrIndx + ";gain=" + result );

		return result;
	}

	/**
	 * Calculates subset entropy where given attribute is a specific value.
	 * 
	 * @param dataSet Data set to be analyzed
	 * @param attrIndx Attribute index to be measured
	 * @param value Attribute value to be filtered on
	 * @return Entropy value for the subset
	 */
	public Double subsetEntropy( DataSet dataSet, Integer attrIndx, DataAttribute value ) {

		log.trace( "Calculating subset entropy for index=" + attrIndx + ";value=" + value );

		DataSet newData = DataSetFactory.emptyDataSet( dataSet );

		List<DataEntry> filtered = new ArrayList<>();

		// Find subset of entries where attribute is given value
		for( DataEntry entry : dataSet.getData() ) {

			log.trace( "Filtering example with value=" + entry.getAttrValue( attrIndx ) );

			if( value.compareTo( entry.getAttrValue( attrIndx ) ) == 0 ) {

				log.trace( "Adding example to filtered dataset." );

				filtered.add( entry );
			}
		}

		newData.setData( filtered );
		newData.setHeader( dataSet.getHeader() );

		// Entropy(Sv)
		Double result = ent.calcEntropy( newData );

		// (|Sv| / |S|) * Entropy(Sv)
		result = ( (double) filtered.size() / dataSet.getData().size() ) * result;

		log.trace( "Subset entropy=" + result );

		return result;
	}
}
