package ee.ut.joshi.connectsix.ai.randomforest.train;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

import ee.ut.joshi.connectsix.ai.randomforest.dataattr.DataAttribute;
import ee.ut.joshi.connectsix.ai.randomforest.dataattr.DataEntry;
import ee.ut.joshi.connectsix.ai.randomforest.dataattr.DataSet;
import ee.ut.joshi.connectsix.ai.randomforest.db.DataSetFactory;
import ee.ut.joshi.connectsix.ai.randomforest.tree.DecisionTree;
import ee.ut.joshi.connectsix.ai.randomforest.tree.TreeLeaf;
import ee.ut.joshi.connectsix.ai.randomforest.tree.TreeNode;
import ee.ut.joshi.connectsix.ai.randomforest.tree.TreeNodeFailure;
import ee.ut.joshi.connectsix.ai.randomforest.tree.TreeRoot;

/**
 * Trainer for a single decision tree based on given data.
 */
public class TrainTree {

	private static final Logger log = Logger.getLogger( TrainTree.class );

	/** Singleton pattern */
	private static TrainTree instance = new TrainTree();

	private TrainGain trainGain = new TrainGain();
	private TrainTreeLeaf trainTreeLeaf = new TrainTreeLeaf();

	private TrainTree() {}

	/**
	 * @return A instance of decision tree trainer
	 */
	public static TrainTree getInstance() {

		return instance;
	}

	/**
	 * Trains a single decision tree using ID3 algorithm.
	 * 
	 * @param trainingSet Data to be trained on
	 * @param attr List of attribute indexes to be used in training
	 * @return Fully trained (not pruned) single decision tree
	 */
	public DecisionTree trainID3tree( DataSet trainingSet, List<Integer> attr ) {

		log.trace( "Starting to train tree" );

		DecisionTree tree = new DecisionTree();

		TreeNode root = trainID3treeRec( trainingSet, attr, null );

		tree.setRoot( root );

		return tree;
	}

	/**
	 * Recursive call for the ID3 training algorithm
	 * 
	 * @param trainingSet Data to be trained on
	 * @param attr List of attribute indexes to be used in training
	 * @param node
	 * @return Returns TreeRoot from first call, otherwise a subset of the tree
	 */
	private TreeNode trainID3treeRec( DataSet trainingSet, List<Integer> attr, TreeNode node ) {

		log.trace( "Recursive train tree call" );

		if( node == null ) {
			node = new TreeRoot();
		}

		// In case we have no more data to analyze
		TreeNode result = trainTreeLeaf.trainID3leaf( trainingSet, attr, node );

		if( result instanceof TreeLeaf ) {

			log.trace( result );

			return result;

		} else if( result instanceof TreeNodeFailure ) {

			return result;
		}

		// Find attribute with best gain and its index
		Integer bestIndx = findBestGain( trainingSet, attr );

		node.setHeaderIndx( bestIndx );

		// THEORY
		// Let D be the attribute with largest Gain(D,S)
		// among attributes in R;
		// Let {dj| j=1,2, .., m} be the values of attribute D;
		// Let {Sj| j=1,2, .., m} be the subsets of S consisting
		// respectively of records with value dj for attribute D;
		// Return a tree with root labeled D and arcs labeled
		// d1, d2, .., dm going respectively to the trees

		log.trace( "Constructing datasets per category" );

		Map<DataAttribute, DataSet> dataSets = new HashMap<>();

		// Create data sets per each category value
		for( DataEntry dataEntry : trainingSet.getData() ) {

			DataAttribute attrVal = dataEntry.getAttrValue( bestIndx );

			if( !dataSets.containsKey( attrVal ) ) {

				log.trace( "Found new attribute value, attrVal=" + attrVal );

				DataSet dataSet = DataSetFactory.emptyDataSet( trainingSet );

				dataSet.addDataEntry( dataEntry );

				dataSets.put( attrVal, dataSet );

			} else {

				log.trace( "Adding new dataEntry for attrVal=" + attrVal );

				dataSets.get( attrVal ).addDataEntry( dataEntry );
			}
		}

		// Recursively train tree for each dataset (one per category value)
		for( Map.Entry<DataAttribute, DataSet> mapEntry : dataSets.entrySet() ) {

			TreeNode child = new TreeNode( node );

			List<Integer> cloneAttr = new ArrayList<>();
			cloneAttr.addAll( attr );
			cloneAttr.remove( bestIndx );

			// This one is big time consumer because of key:
			// log.trace( "Recursive call would be:" + mapEntry.getValue() + "\n" + cloneAttr );

			child = trainID3treeRec( mapEntry.getValue(), cloneAttr, child );

			child.setBranchValue( mapEntry.getKey() );
			node.addNode( child );
		}

		return node;
	}

	/**
	 * Find the attribute index with best information gain.
	 * 
	 * @param trainingSet Data set to be analyzed
	 * @param attr List of attribute indexes to be used in training
	 * @return Index of attribute with highest information gain
	 */
	private Integer findBestGain( DataSet trainingSet, List<Integer> attr ) {

		Integer bestIndx = -1;
		Double bestGain = -1d;

		// Find attribute with most gain
		for( Integer attrIndx : attr ) {

			Double gain = trainGain.calculateGain( trainingSet, attrIndx );

			log.trace( "Calculated gain for attrIndx=" + attrIndx + ";gain=" + gain );

			if( gain > bestGain ) {

				bestGain = gain;
				bestIndx = attrIndx;
			}
		}

		log.trace( "Best gain found for bestIndx=" + bestIndx + ";bestGain=" + bestGain );

		return bestIndx;
	}
}
