package ee.ut.joshi.connectsix.ai.randomforest.train;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

import ee.ut.joshi.connectsix.ai.randomforest.dataattr.DataAttribute;
import ee.ut.joshi.connectsix.ai.randomforest.dataattr.DataEntry;
import ee.ut.joshi.connectsix.ai.randomforest.dataattr.DataSet;
import ee.ut.joshi.connectsix.ai.randomforest.tree.TreeLeaf;
import ee.ut.joshi.connectsix.ai.randomforest.tree.TreeNode;
import ee.ut.joshi.connectsix.ai.randomforest.tree.TreeNodeFailure;

public class TrainTreeLeaf {

	private static final Logger log = Logger.getLogger( TrainTreeLeaf.class );

	/**
	 * Checks if training process has found a decision tree leaf, i.e. the node
	 * that can determine data entry category
	 * 
	 * @param trainingSet Data set to be analyzed
	 * @param attr List of attribute indexes to be used in training
	 * @param node Returns does not affect input node if leaf node is not found
	 * @return Returns same as unaffected input node or a new #TreeLeaf
	 */
	protected TreeNode trainID3leaf( DataSet trainingSet, List<Integer> attr, TreeNode node ) {

		TreeNode temp;

		// If data set size 0
		temp = checkDataSet( trainingSet, node );

		if( temp instanceof TreeNodeFailure ) {

			return temp;
		}

		// If all categories in the data set are identical
		temp = checkCategories( trainingSet, node );

		if( temp instanceof TreeLeaf ) {

			return temp;
		}

		// If there is no more attributes left to be trained on
		temp = checkAttr( trainingSet, attr, node );

		if( temp instanceof TreeLeaf ) {

			return temp;
		}

		return node;
	}

	private TreeNode checkDataSet( DataSet trainingSet, TreeNode node ) {

		if( trainingSet.getData().size() == 0 ) {

			log.error( "No more data to train on, returning failure node" );

			TreeNodeFailure fail = new TreeNodeFailure( node.getParent() );

			fail.setBranchValue( node.getBranchValue() );
			fail.setHeaderIndx( node.getHeaderIndx() );

			node = fail;

			return fail;
		}

		return node;
	}

	// In case there is only one category left
	private TreeNode checkCategories( DataSet trainingSet, TreeNode node ) {

		DataAttribute firstCat = trainingSet.getData().get( 0 ).getCategory();
		DataAttribute tempCat = firstCat;

		for( DataEntry dataEntry : trainingSet.getData() ) {

			if( firstCat.compareTo( dataEntry.getCategory() ) != 0 )
				tempCat = dataEntry.getCategory();
		}

		if( firstCat.compareTo( tempCat ) == 0 ) {

			log.trace( "Only one unique category, constructing TreeLeaf;category=" + firstCat );

			TreeLeaf treeLeaf = new TreeLeaf( node.getParent(), firstCat );

			treeLeaf.setBranchValue( node.getBranchValue() );
			treeLeaf.setHeaderIndx( node.getHeaderIndx() );

			node = treeLeaf;

			return treeLeaf;
		}

		return node;
	}

	// In case we run out of attributes to analyze
	private TreeNode checkAttr( DataSet trainingSet, List<Integer> attr, TreeNode node ) {

		if( attr.size() == 0 ) {

			log.trace( "Number of attributes is zero" );

			DataAttribute freqCat = mostFrequentCategory( trainingSet );

			TreeLeaf treeLeaf = new TreeLeaf( node.getParent(), freqCat );

			treeLeaf.setBranchValue( node.getBranchValue() );
			treeLeaf.setHeaderIndx( node.getHeaderIndx() );

			node = treeLeaf;

			return treeLeaf;
		}

		return node;
	}

	/**
	 * Finds most frequent category in given data set.
	 * 
	 * @param trainingSet Data set to be analyzed
	 * @return Most frequent category value in data set
	 */
	private DataAttribute mostFrequentCategory( DataSet trainingSet ) {

		log.trace( "Looking for most frequent category" );

		HashMap<DataAttribute, Integer> map = new HashMap<>();

		// Find frequency for all categories
		for( DataEntry entry : trainingSet.getData() ) {

			DataAttribute category = entry.getCategory();

			if( map.containsKey( category ) ) {

				Integer count = map.get( category );

				map.put( category, count + 1 );

			} else {

				map.put( category, 1 );
			}
		}

		Set<Map.Entry<DataAttribute, Integer>> mapSet = map.entrySet();

		Map.Entry<DataAttribute, Integer> better = mapSet.iterator().next();

		// Find most frequent attribute
		for( Map.Entry<DataAttribute, Integer> entry : mapSet ) {

			if( entry.getValue() > better.getValue() )
				better = entry;
		}

		log.trace( "Most frequent category=" + better.getKey() + ";frequency=" + better.getValue() );

		return better.getKey();
	}
}
