package ee.ut.joshi.connectsix.ai.randomforest.train;

import java.util.List;

import org.apache.log4j.Logger;

import ee.ut.joshi.connectsix.ai.randomforest.CachedRandomWoods;
import ee.ut.joshi.connectsix.ai.randomforest.ForestTree;
import ee.ut.joshi.connectsix.ai.randomforest.RandomWoods;
import ee.ut.joshi.connectsix.ai.randomforest.TreeReplicator;
import ee.ut.joshi.connectsix.ai.randomforest.dataattr.DataSet;
import ee.ut.joshi.connectsix.ai.randomforest.db.DataSetDao;
import ee.ut.joshi.connectsix.ai.randomforest.tree.DecisionTree;
import ee.ut.joshi.connectsix.main.ConnectObotProps;

public class TrainWoods {

	private static final Logger log = Logger.getLogger( TrainWoods.class );

	/** Determines if a cached or replicated tree will be trained */
	private final static boolean trainCached = ConnectObotProps.forestTrainCached();

	/** A single decision tree trainer */
	private TrainTree trainTree = TrainTree.getInstance();

	/** Database connection to get the data information */
	private DataSetDao dataSetDao;

	public TrainWoods( DataSetDao dataSetDao ) {

		this.dataSetDao = dataSetDao;
	}

	/**
	 * Trains full forest using DB connection from dataSetDao.
	 * 
	 * @return Fully trained forest
	 */
	public RandomWoods trainWoods() {

		DataSet dataSet = dataSetDao.getDataSet();

		return trainWoods( dataSet );
	}

	/**
	 * Trains a single ForestTree based on given input data set.
	 * 
	 * @param dataSet Training data for a single tree
	 * @return Fully trained forest tree
	 */
	private ForestTree trainForestTree( DataSet dataSet, List<Integer> trainingAttr ) {

		DecisionTree tree;

		// Each individual tree is trained here
		tree = trainTree.trainID3tree( dataSet, trainingAttr );

		ForestTree fTree = new ForestTree( tree, trainingAttr );

		return fTree;
	}

	/**
	 * Trains a RandomForest consisting of ForestTrees
	 * 
	 * @param dataSet The training data set
	 * @return Fully trained RandomForest
	 */
	public RandomWoods trainWoods( DataSet dataSet ) {

		// The main training attributes
		List<Integer> baseAttrs = AttributeGenerator.getTrainingAttr();

		return trainWoods( dataSet, baseAttrs );
	}

	/**
	 * Trains either replicated or cached RandomWoods consisting of many ForestTrees or cache<br>
	 * based on provided attribute list.<br>
	 * Can be used to retrain old woods with same patter.
	 * 
	 * @param dataSet The training data set
	 * @return Fully trained RandomForest
	 */
	public RandomWoods trainWoods( DataSet dataSet, List<Integer> baseAttrs ) {

		RandomWoods result;

		if( trainCached ) {

			result = trainCachedWoods( dataSet, baseAttrs );

		} else {

			result = trainReplicatedWoods( dataSet, baseAttrs );
		}

		log.debug( "New woods trained on baseAttrs=" + baseAttrs );

		return result;
	}

	/**
	 * Trains a RandomWoods consisting of ForestTrees based on provided attribute list.<br>
	 * Can be used to retrain old woods with same pattern.
	 * 
	 * @param dataSet The training data set
	 * @return Fully trained RandomForest
	 */
	public RandomWoods trainReplicatedWoods( DataSet dataSet, List<Integer> baseAttrs ) {

		RandomWoods woods = new RandomWoods();

		woods.setBaseAttrs( baseAttrs );

		// All training attributes, including mirrored and transposed/mirrored
		List<List<Integer>> trainingAttr = AttributeGenerator.generateAttributes( baseAttrs );

		// TODO: Smart fields vs regular fields
		List<Integer> smartAttrs = AttributeGenerator.transformToSmart( trainingAttr.get( 0 ) );

		// Main trained tree, not added to woods, but replicated all over the board (also mirrored)
		ForestTree mainTree = trainForestTree( dataSet, smartAttrs );

		TreeReplicator replicator = new TreeReplicator();

		// Replicate trained tree for each attribute subset, e.g. one tree for each mirrored attributes
		for( List<Integer> newAttrs : trainingAttr ) {

			// Replicate tree with same knowledge, but new location on the board
			ForestTree fTree = replicator.replicateTree( mainTree, newAttrs );

			// Add replicated tree to woods
			woods.addTree( fTree );
		}

		return woods;
	}

	/**
	 * Trains a CachedRandomWoods consisting of one ForestTree based on provided attribute list.<br>
	 * Maintains only single decision tree and all other locations/mirroring/rotating is in cache.<br>
	 * Can be used to retrain old woods with same pattern.
	 * 
	 * @param dataSet The training data set
	 * @return Fully trained RandomForest
	 */
	public RandomWoods trainCachedWoods( DataSet dataSet, List<Integer> baseAttrs ) {

		CachedRandomWoods woods = new CachedRandomWoods();

		woods.setBaseAttrs( baseAttrs );

		// All training attributes, including mirrored and transposed/mirrored
		List<List<Integer>> trainingAttr = AttributeGenerator.generateAttributes( baseAttrs );

		// TODO: Smart fields vs regular fields
		baseAttrs = AttributeGenerator.transformToSmart( trainingAttr.get( 0 ) );

		// Main trained tree, not added to woods, but replicated all over the board (also mirrored)
		ForestTree mainTree = trainForestTree( dataSet, baseAttrs );

		woods.setTrainingAttrs( trainingAttr );
		woods.buildCache( mainTree );

		// No need to replicate, just add the main tree. Everything else is in cache
		woods.addTree( mainTree );

		return woods;
	}
}
