package ee.ut.joshi.connectsix.logic;

import java.util.ArrayList;
import java.util.List;
import java.util.Observable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.apache.log4j.Logger;

import ee.ut.joshi.connectsix.ai.gametree.threaded.NegamaxManager;
import ee.ut.joshi.connectsix.logic.board.LinearBoardEvaluator;
import ee.ut.joshi.connectsix.logic.factory.PlayerCloneFactory;
import ee.ut.joshi.connectsix.main.ConnectObotProps;
import ee.ut.joshi.connectsix.model.Board;
import ee.ut.joshi.connectsix.model.Player;
import ee.ut.joshi.connectsix.model.SmartGomokuBoard;
import ee.ut.joshi.connectsix.model.linear.LinearBoard;
import ee.ut.joshi.connectsix.model.player.DrawPlayer;

public class Competition extends Observable implements Runnable {

	private static final Logger log = Logger.getLogger( Competition.class );

	// Threading
	private static final int NROFTHREADS = ConnectObotProps.nrOfThreads();

	private final List<Player> players;

	public Competition( List<Player> players ) {

		this.players = players;
	}

	@Override
	public void run() {

		// Initialize and pre-compile patterns
		LinearBoardEvaluator.getInstance();

		// Add NegamaxManager as observer
		addObserver( NegamaxManager.instance() );

		startCompetition();
	}

	public void startCompetition() {

		log.info( "Starting competition" );

		List<GameController> gameList = initGames( players );

		startGames( gameList );

		analyzeGames( players, gameList );

	}

	protected List<GameController> initGames( List<Player> players ) {

		List<GameController> gameList = new ArrayList<>();

		for( Player one : players ) {
			for( Player two : players ) {

				if( one == two )
					continue;

				Player compOne = PlayerCloneFactory.clonePlayer( one );
				Player compTwo = PlayerCloneFactory.clonePlayer( two );

				compOne.setValue( Board.PLRONE );
				compTwo.setValue( Board.PLRTWO );

				LinearBoard board = new SmartGomokuBoard();
				GameController game = new GameController( board, compOne, compTwo );

				game.logAllMoves();
				game.logFinalStates();

				gameList.add( game );
			}
		}

		return gameList;
	}

	private void startGames( List<GameController> gameList ) {

		// Create thread pool for execution
		ExecutorService execSvc = Executors.newFixedThreadPool( NROFTHREADS );

		for( GameController game : gameList ) {

			execSvc.execute( game );
		}

		log.info( "All games added to thread pool, number of games=" + gameList.size() );

		// Make thread pool closed - no new games can be added
		execSvc.shutdown();

		long start = System.currentTimeMillis();

		// Wait until all games have been completed
		while( !execSvc.isTerminated() ) {

			if( System.currentTimeMillis() - start > 1000 * 10 ) {

				start = System.currentTimeMillis();

				logStates( gameList );
			}

			try {
				Thread.sleep( 1000L );
			} catch( InterruptedException e ) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

		Logger.getLogger( "traintimes" ).debug( "\n" );

		log.info( "All games have finished" );
	}

	private void analyzeGames( List<Player> players, List<GameController> gameList ) {

		log.info( "Starting to analyze game" );

		for( GameController game : gameList ) {

			Player plOne = game.getPlayerOne();
			Player plTwo = game.getPlayerTwo();

			Player winner = game.getWinner();

			log.info( "Game between plOne=" + plOne.getName() + ";plTwo=" + plTwo.getName() + ";winner=" + winner.getName() );
		}

		String header = "\t   ";
		for( Player player : players ) {

			header += "\t" + player.getName();
		}

		log.info( header );

		int midIndx = 0;
		for( Player player : players ) {

			String plrSummary = "\t" + player.getName();

			List<GameController> asFirst = new ArrayList<>();
			List<GameController> asSecnd = new ArrayList<>();

			for( GameController game : gameList ) {
				if( game.getPlayerOne().getName().equals( player.getName() ) )
					asFirst.add( game );
				if( game.getPlayerTwo().getName().equals( player.getName() ) )
					asSecnd.add( game );
			}

			for( GameController game : asFirst ) {

				if( midIndx == asFirst.indexOf( game ) )
					plrSummary += "\t 0 ";

				if( game.getWinner().getName().equals( player.getName() ) )
					plrSummary += "\t 2 ";

				if( game.getWinner().getName().equals( game.getPlayerTwo().getName() ) )
					plrSummary += "\t-2 ";

				if( game.getWinner().compareTo( new DrawPlayer() ) == 0 )
					plrSummary += "\t 1 ";
			}

			log.info( plrSummary );

			midIndx++;
		}
	}

	protected void logStates( List<GameController> gameList ) {

		String result = "Games current move count: ";

		Integer runningGames = 0;
		for( GameController game : gameList ) {

			boolean gameEnded = ( game.getWinner() == null ? false : true );

			String winner = "";

			if( gameEnded ) {

				winner += game.getWinner().getValue();

			} else {

				winner += "-";
				runningGames++;
			}

			result += game.getMoveCounter() + ":" + winner + ";";
		}

		log.info( result );

		// Notify MinimaxManager number of currently running games
		setChanged();
		notifyObservers( runningGames );
		clearChanged();
	}
}
