package ee.ut.joshi.connectsix.logic;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.apache.log4j.Logger;

import ee.ut.joshi.connectsix.ai.genetic.Brain;
import ee.ut.joshi.connectsix.ai.genetic.Individual;
import ee.ut.joshi.connectsix.ai.randomforest.connect6.Connect6DataEntry;
import ee.ut.joshi.connectsix.ai.randomforest.dataattr.DataAttributeDraw;
import ee.ut.joshi.connectsix.ai.randomforest.dataattr.DataAttributePlayer;
import ee.ut.joshi.connectsix.ai.randomforest.db.DataEntryDao;
import ee.ut.joshi.connectsix.ai.randomforest.train.TrainForest;
import ee.ut.joshi.connectsix.main.ConnectObotProps;
import ee.ut.joshi.connectsix.model.Connect6Board;
import ee.ut.joshi.connectsix.model.Player;

public class Connect6Trainer extends Thread {

	private static final Logger log = Logger.getLogger( Connect6Trainer.class );

	// In case the player makes the first move
	public static final int WINVALUE = ConnectObotProps.trainerScoreValues()[ 0 ];
	public static final int DRAWVALUE = ConnectObotProps.trainerScoreValues()[ 1 ];;
	public static final int LOSSVALUE = ConnectObotProps.trainerScoreValues()[ 2 ];;

	// In case the player starts as second player
	public static final int WINVALUE2 = ConnectObotProps.trainerScoreValues()[ 3 ];;
	public static final int DRAWVALUE2 = ConnectObotProps.trainerScoreValues()[ 4 ];;
	public static final int LOSSVALUE2 = ConnectObotProps.trainerScoreValues()[ 5 ];;

	private static final int NROFITERS = ConnectObotProps.nrOfIterations();

	private int currIter = 0;

	private final Brain brain;

	private final DataEntryDao entryDao;

	// Threading
	public static final int NROFTHREADS = ConnectObotProps.nrOfThreads();

	public Connect6Trainer( TrainForest trainForest, DataEntryDao entryDao ) {

		super( "trainer" );
		// Prevents execution in user space and will terminate normally
		setDaemon( true );

		brain = new Brain( trainForest );
		brain.initPopulation();

		this.entryDao = entryDao;
	}

	@Override
	public void start() {

		startTraining();

		Thread.yield();
	}

	private void startTraining() {
		log.info( "Starting training." );

		currIter = 0;

		while( currIter++ < NROFITERS ) {

			log.info( "Training iteration=" + currIter );
			log.info( "Starting brain iteration" );

			brain.iteration();

			log.info( "Brain iteration ended" );

			List<Individual> pop = brain.getPopulation();

			List<GeneticGameController> gameList = startGames( pop );

			analyzeGames( gameList );
		}

		log.info( "Iterations full, training complete" );
	}

	private List<GeneticGameController> startGames( List<Individual> pop ) {

		List<GeneticGameController> gameList = new ArrayList<>();

		// Create thread pool for execution
		ExecutorService execSvc = Executors.newFixedThreadPool( NROFTHREADS );
		
		boolean logAllMoves = true;
		boolean logFinalStates = true;

		for( Individual indivOne : pop ) {
			for( Individual indivTwo : pop ) {

				GeneticGameController game = new GeneticGameController( new Connect6Board(), indivOne, indivTwo );
				
				// Log all moves for one game only
				if( logAllMoves ) {

					game.logAllMoves();
					logAllMoves = false;
				}
				

				// Log final states for first individual only
				if( logFinalStates )
					game.logFinalStates();

				gameList.add( game );

				// Add game to thread
				execSvc.execute( game );
			}
			
			logFinalStates = false;
		}

		log.info( "All games added to thread pool" );

		// Make thread pool closed - no new games added
		execSvc.shutdown();

		// Wait until all games have been completed
		while( !execSvc.isTerminated() ) {

			try {
				Thread.sleep( 10L );
			} catch( InterruptedException e ) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

		Logger.getLogger( "traintimes" ).debug( "\n" );

		log.info( "All games have finished" );

		return gameList;
	}

	private void analyzeGames( List<GeneticGameController> gameList ) {

		log.info( "Starting to analyze game" );

		for( GeneticGameController game : gameList ) {

			Individual indivOne = game.getIndivOne();
			Individual indivTwo = game.getIndivTwo();

			Player plOne = game.getPlayerOne();
			Player plTwo = game.getPlayerTwo();

			Player winner = game.startGame();

			DataAttributePlayer winnerAttr;

			log.trace( "Winner is: " + winner );

			if( winner == plOne ) {
				indivOne.setValue( indivOne.getValue() + WINVALUE );
				indivOne.incWinsOne();
				indivTwo.setValue( indivTwo.getValue() + LOSSVALUE2 );
				indivTwo.incLossTwo();

				winnerAttr = new DataAttributePlayer( plOne.getValue() );
			} else if( winner == plTwo ) {
				indivOne.setValue( indivOne.getValue() + LOSSVALUE );
				indivOne.incLossOne();
				indivTwo.setValue( indivTwo.getValue() + WINVALUE2 );
				indivTwo.incWinsTwo();

				winnerAttr = new DataAttributePlayer( plTwo.getValue() );
			} else {
				indivOne.setValue( indivOne.getValue() + DRAWVALUE );
				indivOne.incDrwsOne();
				indivTwo.setValue( indivTwo.getValue() + DRAWVALUE2 );
				indivTwo.incDrwsTwo();

				winnerAttr = new DataAttributeDraw();
			}

			Connect6Board board = (Connect6Board) game.getBoard();

			Connect6DataEntry dataEntry = new Connect6DataEntry();

			dataEntry.setFields( board.getFields() );
			dataEntry.setWinner( winnerAttr );

			entryDao.saveDataEntry( dataEntry );
		}
	}
}
