package ee.ut.joshi.connectsix.logic;

import org.apache.log4j.Logger;

import ee.ut.joshi.connectsix.model.Board;
import ee.ut.joshi.connectsix.model.DualMove;
import ee.ut.joshi.connectsix.model.Move;
import ee.ut.joshi.connectsix.model.Player;
import ee.ut.joshi.connectsix.model.player.DrawPlayer;

/**
 * Controls the game play of a single game between two players.
 */
public class GameController implements Runnable {

	private static final Logger log = Logger.getLogger( GameController.class );

	/** Board where players play */
	private Board<? extends Object> board;

	/** Players */
	private Player playerOne;
	private Player playerTwo;

	/** Current active player */
	private Player activePlayer;

	/** Tracking for Dual.class moves - false if player did already first move */
	private boolean firstMoveDone = false;

	/** Records player who won the game */
	private Player winner = null;

	private int moveCounter = 0;

	private boolean logAllMoves = false;
	private boolean logFinalStates = false;

	public GameController() {}

	public GameController( Board<? extends Object> board, Player one, Player two ) {

		this.board = board;
		playerOne = one;
		playerTwo = two;

		setActivePlayer( one );
	}

	public void logAllMoves() {
		logAllMoves = true;
	}

	public void logFinalStates() {
		logFinalStates = true;
	}

	@Override
	public void run() {

		log.trace( "Starting game between " + playerOne + " and " + playerTwo );

		long start = System.currentTimeMillis();

		try {

			startGame();

		} catch( Exception e ) {

			e.printStackTrace();
			System.exit( 1 );
		}

		long end = System.currentTimeMillis();

		Logger.getLogger( "traintimes" ).trace( ( end - start ) + ";" );

		log.debug( "Finished game between " + playerOne + " and " + playerTwo + ", winner=" + winner );
	}

	/**
	 * Starts a game between two players.
	 * 
	 * @return Winning player
	 */
	public Player startGame() {

		log.trace( "Starting a game" );

		while( !board.isGameEnd() ) {

			log.trace( "Game not ended" );

			Move move = activePlayer.getNextMove( board );

			if( logAllMoves ) {

				makeMove( move );

				Logger.getLogger( "playermoves" ).trace( "Made move for player=" + activePlayer + ";move=" + move );
				Logger.getLogger( "playermoves" ).trace( "New board state=\n" + board );

			} else {

				makeMove( move );
			}

			moveCounter++;
		}

		Byte win = board.getWinningPlayer();

		winner = win.compareTo( playerOne.getValue() ) == 0 ? playerOne : playerTwo;
		winner = win.compareTo( Board.EMPTY ) == 0 ? new DrawPlayer() : winner;

		if( logFinalStates ) {

			Logger.getLogger( "finalstates" ).trace( "Finished game between " + playerOne + " and " + playerTwo + ", winner=" + winner );
			Logger.getLogger( "finalstates" ).trace( "Winning position: " + board.getWinningPosition() );
			Logger.getLogger( "finalstates" ).trace( "Winner board: " + board );
		}

		return winner;
	}

	/**
	 * Makes a move for active player and changes the active player.
	 * 
	 * @param move Move to be made for current active player
	 */
	protected void makeMove( Move move ) {

		board.makeMove( activePlayer, move );

		// Check if first dual move has been already done
		if( Board.MOVETYPE == DualMove.class && firstMoveDone == false ) {

			// If not, then it was just done, add tracking and do not swap active players
			firstMoveDone = true;

			return;

		} else if( Board.MOVETYPE == DualMove.class && firstMoveDone == true ) {

			// Reset tracker and continue swapping active player
			firstMoveDone = false;
		}

		// Swap active player
		activePlayer = ( activePlayer == playerOne ) ? playerTwo : playerOne;
	}

	public Board<? extends Object> getBoard() {
		return board;
	}

	public void setBoard( Board<? extends Object> board ) {
		this.board = board;
	}

	public Player getPlayerOne() {
		return playerOne;
	}

	public void setPlayerOne( Player playerOne ) {
		this.playerOne = playerOne;
	}

	public Player getPlayerTwo() {
		return playerTwo;
	}

	public void setPlayerTwo( Player playerTwo ) {
		this.playerTwo = playerTwo;
	}

	public Player getActivePlayer() {
		return activePlayer;
	}

	public void setActivePlayer( Player activePlayer ) {
		this.activePlayer = activePlayer;
	}

	public Player getWinner() {
		return winner;
	}

	public void setWinner( Player winner ) {
		this.winner = winner;
	}

	public int getMoveCounter() {
		return moveCounter;
	}
}
