package ee.ut.joshi.connectsix.logic;

import java.util.List;
import java.util.Observable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.apache.log4j.Logger;

import ee.ut.joshi.connectsix.ai.gametree.threaded.NegamaxManager;
import ee.ut.joshi.connectsix.ai.genetic.Brain;
import ee.ut.joshi.connectsix.ai.genetic.Individual;
import ee.ut.joshi.connectsix.ai.randomforest.dataattr.DataAttributeDraw;
import ee.ut.joshi.connectsix.ai.randomforest.dataattr.DataAttributePlayer;
import ee.ut.joshi.connectsix.ai.randomforest.db.DataEntryDao;
import ee.ut.joshi.connectsix.ai.randomforest.train.TrainForest;
import ee.ut.joshi.connectsix.logic.board.LinearBoardEvaluator;
import ee.ut.joshi.connectsix.main.ConnectObotProps;
import ee.ut.joshi.connectsix.model.Player;

abstract public class Trainer extends Observable implements Runnable {

	private static final Logger log = Logger.getLogger( Trainer.class );

	// In case the player makes the first move
	public static final int WINVALUE = ConnectObotProps.trainerScoreValues()[ 0 ];
	public static final int DRAWVALUE = ConnectObotProps.trainerScoreValues()[ 1 ];;
	public static final int LOSSVALUE = ConnectObotProps.trainerScoreValues()[ 2 ];;

	// In case the player starts as second player
	public static final int WINVALUE2 = ConnectObotProps.trainerScoreValues()[ 3 ];;
	public static final int DRAWVALUE2 = ConnectObotProps.trainerScoreValues()[ 4 ];;
	public static final int LOSSVALUE2 = ConnectObotProps.trainerScoreValues()[ 5 ];;

	private static final int NROFITERS = ConnectObotProps.nrOfIterations();

	// Threading
	private static final int NROFTHREADS = ConnectObotProps.nrOfThreads();

	private final Brain brain;

	private final DataEntryDao entryDao;

	private Integer currIter = 0;

	public Trainer( TrainForest trainForest, DataEntryDao entryDao ) {

		brain = new Brain( trainForest );
		brain.initPopulation();

		this.entryDao = entryDao;
	}

	abstract protected List<GeneticGameController> initGames( List<Individual> pop );

	abstract protected void saveGameState( DataEntryDao entryDao2, GeneticGameController game, DataAttributePlayer winnerAttr );

	@Override
	public void run() {

		// Initialize and pre-compile patterns
		LinearBoardEvaluator.getInstance();

		// Add NegamaxManager as observer
		addObserver( NegamaxManager.instance() );

		startTraining();
	}

	private void startTraining() {

		log.info( "Starting training." );

		Runtime runtime = Runtime.getRuntime();

		while( currIter++ < NROFITERS ) {

			log.info( "Training iteration=" + currIter );
			log.info( "Starting brain iteration" );

			brain.iteration();

			log.info( "Brain iteration ended" );

			List<Individual> pop = brain.getPopulation();

			List<GeneticGameController> gameList = initGames( pop );

			// Notify NegamaxManager that we starting new round of training
			setChanged();
			notifyObservers( gameList.size() );
			clearChanged();

			startGames( gameList );

			analyzeGames( gameList );

			log.info( "Memory usage (MB): " + ( ( runtime.totalMemory() - runtime.freeMemory() ) / 1000000 ) );
		}

		log.info( "Iterations full, training complete" );
	}

	private void startGames( List<GeneticGameController> gameList ) {

		// Create thread pool for execution
		ExecutorService execSvc = Executors.newFixedThreadPool( NROFTHREADS );

		for( GeneticGameController game : gameList ) {

			execSvc.execute( game );
		}

		log.info( "All games added to thread pool, number of games=" + gameList.size() );

		// Make thread pool closed - no new games can be added
		execSvc.shutdown();

		long start = System.currentTimeMillis();

		// Wait until all games have been completed
		while( !execSvc.isTerminated() ) {

			if( System.currentTimeMillis() - start > 1000 * 60 ) {

				start = System.currentTimeMillis();

				logStates( gameList );
			}

			try {
				Thread.sleep( 1000L );
			} catch( InterruptedException e ) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

		Logger.getLogger( "traintimes" ).debug( "\n" );

		log.info( "All games have finished" );
	}

	private void analyzeGames( List<GeneticGameController> gameList ) {

		log.info( "Starting to analyze game" );

		for( GeneticGameController game : gameList ) {

			Individual indivOne = game.getIndivOne();
			Individual indivTwo = game.getIndivTwo();

			Player plOne = game.getPlayerOne();
			Player plTwo = game.getPlayerTwo();

			// TODO: What the hell is this doing here???
			// It does not start new game due to board state, but this is plain wrong
			Player winner = game.startGame();

			DataAttributePlayer winnerAttr;

			log.trace( "Winner is: " + winner );

			if( winner == plOne ) {
				indivOne.setValue( indivOne.getValue() + WINVALUE );
				indivOne.incWinsOne();
				indivTwo.setValue( indivTwo.getValue() + LOSSVALUE2 );
				indivTwo.incLossTwo();

				winnerAttr = new DataAttributePlayer( plOne.getValue() );
			} else if( winner == plTwo ) {
				indivOne.setValue( indivOne.getValue() + LOSSVALUE );
				indivOne.incLossOne();
				indivTwo.setValue( indivTwo.getValue() + WINVALUE2 );
				indivTwo.incWinsTwo();

				winnerAttr = new DataAttributePlayer( plTwo.getValue() );
			} else {
				indivOne.setValue( indivOne.getValue() + DRAWVALUE );
				indivOne.incDrwsOne();
				indivTwo.setValue( indivTwo.getValue() + DRAWVALUE2 );
				indivTwo.incDrwsTwo();

				winnerAttr = new DataAttributeDraw();
			}

			saveGameState( entryDao, game, winnerAttr );
		}
	}

	protected void logStates( List<GeneticGameController> gameList ) {

		String result = "Games current move count: ";

		Integer runningGames = 0;
		for( GeneticGameController game : gameList ) {

			boolean gameEnded = ( game.getWinner() == null ? false : true );

			String winner = "";

			if( gameEnded ) {

				winner += game.getWinner().getValue();

			} else {

				winner += "-";
				runningGames++;
			}

			result += game.getMoveCounter() + ":" + winner + ";";
		}

		log.info( result );

		// Notify MinimaxManager number of currently running games
		setChanged();
		notifyObservers( runningGames );
		clearChanged();
	}
}
