package ee.ut.joshi.connectsix.logic.factory;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import ee.ut.joshi.connectsix.model.Board;
import ee.ut.joshi.connectsix.model.CachedC6Board;
import ee.ut.joshi.connectsix.model.CachedGomokuBoard;
import ee.ut.joshi.connectsix.model.ConnectFourBoard;
import ee.ut.joshi.connectsix.model.Move;
import ee.ut.joshi.connectsix.model.SmartC6Board;
import ee.ut.joshi.connectsix.model.SmartGomokuBoard;
import ee.ut.joshi.connectsix.model.TicBoard;
import ee.ut.joshi.connectsix.model.linear.LinearBoard;
import ee.ut.joshi.connectsix.model.matrix.MatrixBoard;

/**
 * Game board cloner - creates new instance of the game board with exact same board state.<br>
 * Used in #MiniMax game tree search algorithm.
 */
public class BoardCloneFactory {

	private static final Logger log = Logger.getLogger( BoardCloneFactory.class );

	/**
	 * Creates new instance of the game board with exact same board state.
	 * 
	 * @param board Game board to be cloned
	 * @return New instance of game board with same state
	 */
	public static Board<? extends Object> cloneBoard( Board<? extends Object> board ) {

		// TODO: Add also lastmove cloning for all boards

		if( board instanceof TicBoard )
			return cloneTicBoard( (TicBoard) board );
		else if( board instanceof CachedC6Board )
			return cloneCachedC6Board( (CachedC6Board) board );
		else if( board instanceof SmartC6Board )
			return cloneSmartC6Board( (SmartC6Board) board );
		else if( board instanceof SmartGomokuBoard )
			return cloneSmartGomokuBoard( (SmartGomokuBoard) board );
		else if( board instanceof CachedGomokuBoard )
			return cloneCachedGomokuBoard( (CachedGomokuBoard) board );
		else if( board instanceof ConnectFourBoard )
			return cloneCachedGomokuBoard( (ConnectFourBoard) board );
		else if( board instanceof MatrixBoard )
			return cloneMatrixBoard( (MatrixBoard) board );
		else if( board instanceof LinearBoard )
			return cloneLinearBoard( (LinearBoard) board );

		log.fatal( "Board cloning not supported!" );
		System.exit( 1 );

		return null;
	}

	private static TicBoard cloneTicBoard( TicBoard board ) {

		TicBoard clone = new TicBoard();

		byte[] fields = board.getFields();
		byte[] cloneFields = new byte[ Board.NROFFIELDS ];

		System.arraycopy( fields, 0, cloneFields, 0, Board.NROFFIELDS );

		clone.setFields( cloneFields );

		return clone;
	}

	private static SmartC6Board cloneSmartC6Board( SmartC6Board board ) {

		SmartC6Board clone = new SmartC6Board( false );

		byte[] fields = board.getFields();
		byte[] cloneFields = new byte[ Board.NROFFIELDS ];

		System.arraycopy( fields, 0, cloneFields, 0, Board.NROFFIELDS );

		clone.setFields( cloneFields );

		List<Move> smartMoves = new ArrayList<>( board.getSmartMoves() );

		clone.setSmartMoves( smartMoves );
		clone.setLastMove( board.getLastMove() );

		return clone;
	}

	private static ConnectFourBoard cloneCachedGomokuBoard( ConnectFourBoard board ) {

		ConnectFourBoard clone = new ConnectFourBoard();

		byte[] fields = board.getFields();
		byte[] cloneFields = new byte[ Board.NROFFIELDS ];

		System.arraycopy( fields, 0, cloneFields, 0, Board.NROFFIELDS );

		clone.setFields( cloneFields );

		return clone;
	}

	private static CachedC6Board cloneCachedC6Board( CachedC6Board board ) {

		CachedC6Board clone = new CachedC6Board( false );

		byte[] fields = board.getFields();
		byte[] cloneFields = new byte[Board.NROFFIELDS];

		System.arraycopy( fields, 0, cloneFields, 0, Board.NROFFIELDS );

		clone.setFields( cloneFields );

		List<Move> cachedMoves = new ArrayList<>( board.getCachedMoves() );

		clone.setCachedMoves( cachedMoves );

		return clone;
	}

	private static SmartGomokuBoard cloneSmartGomokuBoard( SmartGomokuBoard board ) {

		SmartGomokuBoard clone = new SmartGomokuBoard( false );

		byte[] fields = board.getFields();
		byte[] cloneFields = new byte[Board.NROFFIELDS];

		System.arraycopy( fields, 0, cloneFields, 0, Board.NROFFIELDS );

		clone.setFields( cloneFields );

		List<Move> smartMoves = new ArrayList<>( board.getSmartMoves() );

		clone.setSmartMoves( smartMoves );
		clone.setLastMove( board.getLastMove() );

		return clone;
	}

	private static CachedGomokuBoard cloneCachedGomokuBoard( CachedGomokuBoard board ) {

		CachedGomokuBoard clone = new CachedGomokuBoard( false );

		byte[] fields = board.getFields();
		byte[] cloneFields = new byte[Board.NROFFIELDS];

		System.arraycopy( fields, 0, cloneFields, 0, Board.NROFFIELDS );

		clone.setFields( cloneFields );

		List<Move> cachedMoves = new ArrayList<>( board.getCachedMoves() );

		clone.setCachedMoves( cachedMoves );

		return clone;
	}

	private static LinearBoard cloneLinearBoard( LinearBoard board ) {

		LinearBoard clone = new LinearBoard();

		byte[] fields = board.getFields();
		byte[] cloneFields = new byte[Board.NROFFIELDS];

		System.arraycopy( fields, 0, cloneFields, 0, Board.NROFFIELDS );

		clone.setFields( cloneFields );

		return clone;
	}

	private static MatrixBoard cloneMatrixBoard( MatrixBoard board ) {

		MatrixBoard clone = new MatrixBoard();

		byte[][] fields = board.getFields();
		byte[][] clonefields = new byte[Board.WIDTH][Board.WIDTH];

		for( int i = 0; i < Board.WIDTH; i++ ) {

			// Works faster than .clone with small lengths and primitive values.
			System.arraycopy( fields[ i ], 0, clonefields[ i ], 0, Board.WIDTH );
		}

		clone.setFields( clonefields );

		return clone;
	}
}
