package ee.ut.joshi.connectsix.model;

import java.util.List;

import ee.ut.joshi.connectsix.main.ConnectObotProps;

/**
 * Game board interface for various game boards.
 * 
 * @param <T> Game board field type, e.g. single array, double array, list
 */
public interface Board<T extends Object> {

	/** Game field value if it's empty */
	public static final byte EMPTY = 0;

	/** Game field value if player one has made a move there */
	public static final byte PLRONE = 1;

	/** Game field value if player two has made a move there */
	public static final byte PLRTWO = 2;

	/** Game board width as in number of fields */
	public static final int WIDTH = ConnectObotProps.boardWidth();

	/** Number of moves needed to be in a row to make a winning situation */
	public static final int INAROW = ConnectObotProps.boardInaRow();

	/** The move type for the board - e.g. Move or DualMove */
	public static final Class<? extends Move> MOVETYPE = ConnectObotProps.boardMoveType();

	/** Total number of fields on game board */
	public static final int NROFFIELDS = WIDTH * WIDTH;

	/** Number of recursive calls for determining winning situation */
	public static final int RECURSION = INAROW - 1;

	/**
	 * Checks if move is with legal position (e.g. the field is empty)
	 * 
	 * @param move Move to be checked
	 * @return True if provided move is legal, false otherwise
	 */
	public boolean isLegalMove( Move move );

	/**
	 * Makes a move for the given player at given position.
	 * 
	 * @param player Player for who to make the move
	 * @param move Move to be made for given player
	 */
	public void makeMove( Player player, Move move );

	/**
	 * @return Returns last move made on the board
	 */
	public Move getLastMove();


	/**
	 * @param lastMove Stores last move done on the board, should be used with care
	 */
	void setLastMove( Move lastMove );
	
	/**
	 * @return List of all possible moves on current board state
	 */
	public List<Move> getNextMoves();

	/**
	 * @return Determines if there is a winning/draw situation on the board
	 */
	public boolean isGameEnd();

	/**
	 * @return Position where the winning position starts if game has ended
	 */
	public Position getWinningPosition();

	/**
	 * @return Byte value of the winning position.
	 */
	byte getWinningPlayer();

	/**
	 * @return Raw data of game fields
	 */
	public T getFields();

	public void setFields( T fields );
}
