package ee.ut.joshi.connectsix.model.linear;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import ee.ut.joshi.connectsix.logic.board.LinearBoardEvaluator;
import ee.ut.joshi.connectsix.model.Board;
import ee.ut.joshi.connectsix.model.DualMove;
import ee.ut.joshi.connectsix.model.Move;
import ee.ut.joshi.connectsix.model.Player;
import ee.ut.joshi.connectsix.model.Position;

public class LinearBoard implements Board<byte[]> {

	private static final Logger log = Logger.getLogger( LinearBoard.class );

	/** Direction constants for recursive calls */
	private static final byte HORIZONTAL = 1;
	private static final byte VERTICAL = 2;
	private static final byte DIAGONAL = 4;
	private static final byte REVDIAGONAL = 8;

	/** Raw data of the game board fields */
	private byte fields[] = new byte[ NROFFIELDS ];

	/** Keeps track first field of the winning area */
	private Position winningPosition;

	/** Keeps track of byte value on winning position */
	private byte winningPlayer;

	/** Tracks last move made by a player */
	private Move lastMove;

	/** Keeps track of current board score, should be reset on each move */
	private Integer score;

	@Override
	public boolean isLegalMove( Move move ) {

		if( move instanceof LinearMove ) {

			LinearMove linMove = (LinearMove) move;

			LinearPosition linPos = linMove.getPosition();

			if( fields[ linPos.getPosition() ] == Board.EMPTY )
				return true;
		}

		return false;
	}

	@Override
	public void makeMove( Player player, Move move ) {

		// Invalidate current board score since the board state has changed.
		score = null;

		log.trace( "Making move for player " + player + " : " + move );

		try {

			if( move instanceof LinearDualMove ) {

				LinearDualMove linearMove = (LinearDualMove) move;

				fields[ linearMove.getFstPos().getPosition() ] = player.getValue();
				fields[ linearMove.getSndPos().getPosition() ] = player.getValue();

			} else if( move instanceof LinearMove ) {

				LinearMove linearMove = (LinearMove) move;

				fields[ linearMove.getPosition().getPosition() ] = player.getValue();

			} else {

				throw new Exception( "Move not supported." );
			}

		} catch( Exception e ) {

			log.fatal( e.getMessage() );
			e.printStackTrace();

			System.exit( 1 );
		}

		lastMove = move;
	}

	@Override
	public List<Move> getNextMoves() {

		// If game play is with dual moves, e.g. ConnectSix
		if( Board.MOVETYPE == DualMove.class ) {

			return getNextDualMoves();
		}

		List<Move> moves = new ArrayList<Move>();

		// Scan through all positions
		for( int position = 0; position < NROFFIELDS; position++ ) {

			// Check if position is legal move
			if( fields[ position ] == 0 ) {

				LinearPosition fstPos = new LinearPosition( position );

				// Create new legal move and add to list
				moves.add( new LinearMove( fstPos ) );
			}
		}

		return moves;
	}

	public List<Move> getNextDualMoves() {

		List<Move> moves = new ArrayList<Move>();

		// First loop
		for( int positionOne = 0; positionOne < NROFFIELDS; positionOne++ ) {

			// Check if first position is legal move
			if( fields[ positionOne ] == 0 ) {

				// Second loop starting from next position
				for( int positionTwo = positionOne + 1; positionTwo < NROFFIELDS; positionTwo++ ) {

					// Check if second position is legal move
					if( fields[ positionTwo ] == 0 ) {

						LinearPosition fstPos = new LinearPosition( positionOne );
						LinearPosition sndPos = new LinearPosition( positionTwo );

						// Create new legal move and add to list
						moves.add( new LinearDualMove( fstPos, sndPos ) );
					}
				}
			}
		}

		return moves;
	}

	@Override
	public boolean isGameEnd() {

		boolean result = false;

		for( int position = 0; position < NROFFIELDS; position++ ) {

			result = ( result || checkPosition( position, HORIZONTAL ) );
			result = ( result || checkPosition( position, VERTICAL ) );
			result = ( result || checkPosition( position, DIAGONAL ) );
			result = ( result || checkPosition( position, REVDIAGONAL ) );

			if( result ) {

				winningPosition = new LinearPosition( position );
				winningPlayer = fields[ position ];

				return true;
			}
		}

		// Check for draw, e.g. no moves left
		if( getNextMoves().size() == 0 ) {

			result = true;
			winningPosition = new LinearPosition( NROFFIELDS );
			winningPlayer = Board.EMPTY;
		}

		return result;
	}

	private boolean checkPosition( int position, byte direction ) {

		// NOTE: Other way to implement this is one method for each direction,
		// but it has a lot of code duplication (e.g. MatrixBoard)
		// Also.. you know that most math defects are in this method.

		// Positions in last 5 'columns' should not be checked
		// For example with 19*19 board there are 361 positions
		// In this case, if modulo of target position is larger
		// than 13, then this position falls under last 5 columns.
		// For example position '14' as follows:
		// 14 % 19 = 14
		// 14 > (19 - 6)
		// returns true, so the position is in last 5 columns
		if( direction == HORIZONTAL && ( position % WIDTH ) > WIDTH - INAROW )
			return false;

		// We do not need to check positions in bottom 5 lines
		if( direction == VERTICAL && ( position >= WIDTH * ( WIDTH - RECURSION ) ) )
			return false;

		// We do not need to check positions in last 5 columns and in last
		// bottom 5 lines
		if( direction == DIAGONAL )
			if( ( position % WIDTH ) > WIDTH - INAROW || position >= WIDTH * ( WIDTH - RECURSION ) )
				return false;

		// We do not need to check positions in first 5 columns and in last
		// bottom 5 lines
		if( direction == REVDIAGONAL )
			if( ( position % WIDTH ) < RECURSION || position >= WIDTH * ( WIDTH - RECURSION ) )
				return false;

		// Check if first player has winning condition
		if( checkPositionRec( PLRONE, RECURSION, direction, position ) )
			return true;

		// Next check if second player has winning condition
		if( checkPositionRec( PLRTWO, RECURSION, direction, position ) )
			return true;

		return false;
	}

	private boolean checkPositionRec( byte player, int recursion, byte direction, int position ) {

		// If we find a field that is not for player, it's not a winning
		// condition
		if( fields[ position ] != player )
			return false;

		// If we reached max level recursion and all fields so far are player
		// fields then it's a winning condition
		if( recursion == 0 )
			return true;

		if( direction == HORIZONTAL )
			return checkPositionRec( player, --recursion, direction, position + 1 );

		if( direction == VERTICAL )
			return checkPositionRec( player, --recursion, direction, position + WIDTH );

		if( direction == DIAGONAL )
			return checkPositionRec( player, --recursion, direction, position + WIDTH + 1 );

		if( direction == REVDIAGONAL )
			return checkPositionRec( player, --recursion, direction, position + WIDTH - 1 );

		log.error( "Direction byte incorrect!" );

		// We reach this location if direction byte is incorrect
		return false;
	}

	public Integer getBoardScore() {

		if( score == null ) {

			score = LinearBoardEvaluator.getInstance().boardScore( this );
		}

		return score;
	}

	@Override
	public Move getLastMove() {

		return lastMove;
	}

	@Override
	public void setLastMove( Move lastMove ) {

		this.lastMove = lastMove;
	}

	@Override
	public Position getWinningPosition() {

		return winningPosition;
	}

	@Override
	public byte[] getFields() {

		return fields;
	}

	@Override
	public void setFields( byte[] fields ) {

		this.fields = fields;
	}

	@Override
	public byte getWinningPlayer() {

		return winningPlayer;
	}

	@Override
	public String toString() {

		LinearMove linLastMove = (LinearMove) getLastMove();

		int lastpos = -1;

		if( linLastMove != null )
			lastpos = linLastMove.getPosition().getPosition();

		String result = "Board score=" + getBoardScore() + "\n   ";

		for( int i = 0; i < WIDTH; i++ ) {
			result += String.format( "%02d ", i );
		}

		for( int i = 0; i < WIDTH; i++ ) {
			result += String.format( "\n%02d ", i );

			for( int field = i * WIDTH; field < i * WIDTH + WIDTH; field++ ) {

				if( field == lastpos ) {

					result += "## ";
				} else if( fields[ field ] == EMPTY ) {

					result += "-- ";
				} else if( fields[ field ] == PLRONE ) {

					result += "XX ";
				} else if( fields[ field ] == PLRTWO ) {

					result += "OO ";
				}
			}
		}

		return result;
	}

	public String toStringWithMoves() {

		LinearMove linLastMove = (LinearMove) getLastMove();

		int lastpos = -1;

		if( linLastMove != null )
			lastpos = linLastMove.getPosition().getPosition();

		String result = "Board score=" + getBoardScore() + "\n   ";

		for( int i = 0; i < WIDTH; i++ ) {
			result += "   "; // String.format( "%02d ", i );
		}

		for( int i = 0; i < WIDTH; i++ ) {
			result += "\n   "; // String.format( "\n%02d ", i );

			for( int field = i * WIDTH; field < i * WIDTH + WIDTH; field++ ) {

				if( field == lastpos ) {

					result += "## ";
				} else if( fields[ field ] == EMPTY ) {

					result += String.format( "%02d ", field );
				} else if( fields[ field ] == PLRONE ) {

					result += "XX ";
				} else if( fields[ field ] == PLRTWO ) {

					result += "OO ";
				}
			}
		}

		return result;
	}
}
